﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Reflection;
using System.Text;
using Mono.Cecil;
using Mono.Cecil.Cil;
using NETDeob.Deobfuscators;
using NETDeob.Engine.Utils;
using NETDeob.Misc;
using NETDeob.Misc.Structs__Enums___Interfaces.Deobfuscation;
using NETDeob.Misc.Structs__Enums___Interfaces.Tasks.Base;

namespace NETDeob.Deobfuscators.Confuser.Tasks._1_7
{
    class ConstantsDecryptor : DeobfuscationTask
    {
        public static int[] Modifiers;

        public class DecryptionEntry
        {
            public MethodDefinition Caller;

            public int Id;
            public uint MDToken;

            public IEnumerable<Instruction> BadInstructions;

            public string PlainText;
            
        }

        private const int BufSize = 4096;
        private static byte[] _rawData;

        public ConstantsDecryptor(AssemblyDefinition asmDef)
            : base(asmDef)
        {
            RoutineDescription = "Decrypt and replace encrypted constants";
        }

        [DeobfuscationPhase(1, "Locate decryption method")]
        public static bool Phase1()
        {
            var decryptor = AsmDef.FindMethod(IsDecryptor);

            if(decryptor == null){
                ThrowPhaseError("No string encryption?", 0, true);
                return true;
            }

            var resName = decryptor.Body.Instructions.GetOperandAt<string>(OpCodes.Ldstr, 1);

            MarkMember(decryptor);
            MarkMember(decryptor.DeclaringType.Methods.FirstOrDefault(m => m.IsConstructor));

            if(resName == null){
                ThrowPhaseError("Could not locate string resource!", 1, false);
                return false;
            }

            if (DeobfuscatorContext.ResStreams.Count == 0)
            {
                PhaseParam = new object[]
                                 {
                                     Assembly.LoadFile(DeobfuscatorContext.InPath).GetManifestResourceStream(resName),
                                     decryptor
                                 };

                // Resource decryption compability
                DeobfuscatorContext.InPath = DeobfuscatorContext.InPath.Replace("_resdump.exe", null);
            }
            else
                PhaseParam = new object[] { DeobfuscatorContext.ResStreams.First(res => res.Name == resName), decryptor };

            if (PhaseParam[0] == null){
                ThrowPhaseError("Null stream!", 1, false);
                return false;
            }

            Logger.VSLog("Found decryptor method at " + decryptor.Name);

            return true;
        }

        /// <summary>
        /// PhaseParam = { assembly, decryptor }
        /// </summary>
        [DeobfuscationPhase(2, "Decompress resource data")]
        public static bool Phase2()
        {
            var inStream = PhaseParam[0];
            var outStream = new MemoryStream();

            using(var deflater = new DeflateStream(inStream, CompressionMode.Decompress))
            {
                var buf = new byte[BufSize];
                var count = deflater.Read(buf, 0, BufSize);

                while(count != 0)
                {
                    outStream.Write(buf, 0, count);
                    count = deflater.Read(buf, 0, BufSize);
                }
            }

            _rawData = outStream.ToArray();

            if(_rawData.Length == 0)
            {
                ThrowPhaseError("Failed to decompress resource!", 1, false);
                return false;
            }

            Logger.VSLog(string.Format("Decompressed constants resource ({0} bytes)...", _rawData.Length));

            return true;
        }

        /// <summary>
        /// PhaseParam = { assembly, decryptor }
        /// </summary>
        [DeobfuscationPhase(3, "Construct decryption entries")]
        public static bool Phase3()
        {
            // brainfart.. :3
            var decEntries = new List<DecryptionEntry>();
            
            foreach (var mDef in AsmDef.FindMethods(m => true).Where(mDef => !IsDecryptor(mDef)))
            {
                if (!mDef.HasBody)
                    continue;

                for (var i = 0; i < mDef.Body.Instructions.Count; i++)
                {
                    var instr = mDef.Body.Instructions[i];

                    if (instr.Previous == null)
                        continue;

                    if (instr.OpCode != OpCodes.Call || instr.Previous.OpCode != OpCodes.Ldc_I4 ||
                        !IsDecryptor((instr.Operand as MethodReference).Resolve())) continue;

                    var entry = new DecryptionEntry
                                    {
                                        Id = (int) instr.Previous.Operand,
                                        Caller = mDef,
                                        MDToken = mDef.MetadataToken.ToUInt32(),
                                        BadInstructions = mDef.Body.Instructions.GetInstructionBlock(i -1, 3),
                                    };

                    decEntries.Add(entry);
                }
            }

            PhaseParam = new object[] {decEntries, PhaseParam[1]};

            Logger.VSLog(string.Format("{0} decryption entries constructed...", decEntries.Count));
            return true;
        }

        /// <summary>
        /// PhaseParam = { decEntries, decryptor }
        /// </summary>
        [DeobfuscationPhase(4, "Decrypt constants")]
        public static bool Phase4()
        {
            var decEntries = PhaseParam[0];
            InitializeDecryption(PhaseParam[1]);

            foreach (var entry in decEntries)
            {
                var old = string.Format("Id({0})", entry.Id);
                entry.PlainText = DecryptString(entry);

                Logger.VLog(string.Format(@"[Decrypt] ""{0}"" -> ""{1}""", old, entry.PlainText));
            }

            Logger.VSLog(string.Format("{0} constants was decrypted...", decEntries.Count));

            PhaseParam = decEntries;
            return true;
        }

        /// <summary>
        /// PhaseParam = decEntries
        /// </summary>
        [DeobfuscationPhase(5, "Replace constants")]
        public static bool Phase5()
        {
            var decEntries = PhaseParam as List<DecryptionEntry>;

            foreach (var entry in decEntries)
            {
                var ilProc = entry.Caller.Body.GetILProcessor();
                var badInstructions = entry.BadInstructions.ToArray();

                ilProc.InsertAfter(badInstructions[2], ilProc.Create(OpCodes.Ldstr, entry.PlainText));

                foreach (var instr in badInstructions)
                    ilProc.Remove(instr);
            }

            Logger.VSLog("All encrypted constants was replaced with plaintext...");

            return true;
        }

        public static bool IsDecryptor(MethodDefinition mDef)
        {
            if (!mDef.HasBody)
                return false;

            if (!mDef.HasParameters || !mDef.Body.HasVariables)
                return false;

            if (mDef.Body.Variables.Count != 20)
                return false;

            if (mDef.Parameters.Count != 1)
                return false;

            return true;
        }
        public static void InitializeDecryption(MethodDefinition mDef)
        {
            Modifiers = new int[4];

            Modifiers[0] = mDef.Body.Instructions.GetOperandAt<int>(OpCodes.Ldc_I4, 3);
            Modifiers[1] = mDef.Body.Instructions.GetOperandAt<int>(OpCodes.Ldc_I4, 4);
            Modifiers[2] = mDef.Body.Instructions.GetOperandAt<int>(OpCodes.Ldc_I4, 5);
            Modifiers[3] = mDef.Body.Instructions.GetOperandAt<int>(OpCodes.Ldc_I4, 12);
        }

        #region Reversed Methods

        //Copyright (C) 2011 Ki

        //This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
        //This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
        //You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
        //Linking Confuser statically or dynamically with other modules is making a combined work based on Confuser. Thus, the terms and conditions of the GNU General Public License cover the whole combination.
        //In addition, as a special exception, the copyright holders of Confuser give you permission to combine a portion of Confuser with other binary code ("Program") under any license the copyright holders of Program specified, provided the combined work is produced by Confuser.

        static string DecryptString(DecryptionEntry entry)
        {
            object obj2 = null;
            uint num3 = (uint)Modifiers[0] ^ entry.MDToken;
            uint num4 = (uint)Modifiers[1];
            uint num5 = (uint)Modifiers[2];

            for (uint i = 1; i <= 64; i++)
            {
                num3 = (uint)(((num3 & 16777215) << 8) | ((num3 & -16777216) >> 24));
                uint num7 = (num3 & 255) % 64;
                if ((num7 >= 0) && (num7 < 16))
                {
                    num4 |= (((num3 & 65280) >> 8) & ((num3 & 16711680) >> 16)) ^ (~num3 & 255);
                    num5 ^= ((num3 * i) + 1) % 16;
                    num3 += (num4 | num5) ^ (uint)Modifiers[3];
                }
                else if ((num7 >= 16) && (num7 < 32))
                {
                    num4 ^= ((num3 & 16711935) << 8) ^ (((num3 & 16776960) >> 8) | (~num3 & 65535));
                    num5 += (num3 * i) % 32;
                    num3 |= (num4 + ~num5) & (uint)Modifiers[3];
                }
                else if ((num7 >= 32) && (num7 < 48))
                {
                    num4 += ((num3 & 255) | ((num3 & 16711680) >> 16)) + (~num3 & 255);
                    num5 -= ~(num3 + num7) % 48;
                    num3 ^= (num4 % num5) | (uint)Modifiers[3];
                }
                else if ((num7 >= 48) && (num7 < 64))
                {
                    num4 ^= (((num3 & 16711680) >> 16) | ~(num3 & 255)) * (~num3 & 16711680);
                    num5 += (num3 ^ (i - 1)) % num7;
                    num3 -= ~(num4 ^ num5) + (uint)Modifiers[3];
                }
            }

            // I think this is constant
            uint key = num3 ^ (uint)entry.Id;

            using (BinaryReader reader = new BinaryReader(new MemoryStream(_rawData)))
            {
                reader.BaseStream.Seek(key, SeekOrigin.Begin);

                var num9 = reader.ReadByte();
                var bytes = reader.ReadBytes(reader.ReadInt32());

                var random = new Random(Modifiers[3] ^ ((int)key));
                var buffer3 = new byte[bytes.Length];

                random.NextBytes(buffer3);

                BitArray array = new BitArray(bytes);
                array.Xor(new BitArray(buffer3));
                array.CopyTo(bytes, 0);

                obj2 = Encoding.UTF8.GetString(bytes);
            }

            return obj2 as string; // 1.7 doesn't seem to encrypt other constants than strings???
        }

        #endregion
    }
}
